#ifndef WINBASE_H_INC
#define WINBASE_H_INC

#include <windows.h>
#include <stdio.h>
#include <map>

#include "drawstd.h"

class WindowBaseClass
{
public:
   virtual ~WindowBaseClass()
   {
      WindowMapClass::iterator idx = WindowMap.find(WindowHandle);
      if(idx != WindowMap.end())
      {
         WindowMap.erase(idx);
         return;
      }
   }
   HWND GetWindowHandle()
   {
      return WindowHandle;
   }
   virtual void MoveWindow(RECT &aRect, bool ReDraw = true)
   {
      ::MoveWindow(WindowHandle, aRect.left, aRect.top, aRect.right, aRect.bottom, ReDraw);
   }
protected:
   virtual void OnVScroll(int aCode, int aPos)
   {}
   virtual void OnHScroll(int aCode, int aPos)
   {}
   virtual void OnClose()
   {}
   virtual void OnLButtonDown(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}
   virtual void OnLButtonUp(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}
   virtual void OnRButtonDown(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}
   virtual void OnRButtonUp(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}
   virtual void OnLButtonDoubleClick(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}
   virtual void OnMouseMove(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
   {}

   virtual LRESULT OnCreate(CREATESTRUCT *aCreateStruct)
   {
      return 0;
   }

   virtual bool OnCommand(int aId, HWND ahwndCtl, UINT aCNotifyCode)
   {
      return false; // Not Handled
   }

   virtual LRESULT OnNotify(int32_t aCtrlId, NMHDR *aParam)
   {
      return 0;
   }

   static LRESULT CALLBACK WndProcEntry(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
   {
      WindowMapClass::iterator idx = WindowMap.find(hwnd);
      if(idx == WindowMap.end())
      {
         if(msg == WM_CREATE)
         {
            CREATESTRUCT *CreateStruct = (CREATESTRUCT *)lParam;
            WindowBaseClass *Obj = (WindowBaseClass *)CreateStruct->lpCreateParams;
            WindowMap.insert(std::make_pair(hwnd, Obj));
            Obj->WindowHandle = hwnd;
            return Obj->WindowProc(msg, wParam, lParam);
         }
         else
            return DefWindowProc(hwnd, msg, wParam, lParam);
      }
      if(msg == WM_COMMAND)
      { // First loop thru childs
         for(WindowMapClass::iterator ch = WindowMap.begin(); ch != WindowMap.end(); ch++)
         {
            if(ch->second->GetWindowHandle() != idx->second->GetWindowHandle() &&
               idx->second->GetWindowHandle() == GetParent(ch->second->GetWindowHandle()) &&
               ch->second->WindowProc(msg, wParam, lParam))
            {
               return TRUE;
            }
         }
      }
      return idx->second->WindowProc(msg, wParam, lParam);
   }
   virtual void OnMsgMenuSelect(uint16_t aIndex, uint16_t aFlags, HMENU aMenuHandle)
   {}
   virtual LRESULT OnPaint()
   {
      return TRUE; // Not handled
   }

   virtual void OnSize(uint32_t aType, uint16_t aX, uint16_t aY)
   { }
   virtual LRESULT OnEraseBackground(HDC aDC)
   {
      return 0;
   }

   LRESULT WindowProc(UINT msg, WPARAM wParam, LPARAM lParam)
   {
      switch(msg)
      {
      case WM_COMMAND:
         return OnCommand(LOWORD(wParam), (HWND )lParam, HIWORD(wParam));
      case WM_MENUSELECT:
         OnMsgMenuSelect(LOWORD(wParam), HIWORD(wParam), (HMENU )lParam);
         break;
      case WM_DESTROY:
         OnClose();
         PostQuitMessage(0);
         return TRUE;
      case WM_CREATE:
         return OnCreate((CREATESTRUCT *)lParam);
      case WM_PAINT:
         if(!OnPaint())
            return FALSE;
         break;
      case WM_SIZE:
         OnSize(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_VSCROLL:
         OnVScroll(LOWORD(wParam), HIWORD(wParam));
         break;
      case WM_HSCROLL:
         OnHScroll(LOWORD(wParam), HIWORD(wParam));
         break;
      case WM_ERASEBKGND:
         if(OnEraseBackground((HDC )wParam))
            return TRUE;
         break;
      case WM_LBUTTONDOWN:
         OnLButtonDown(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_LBUTTONUP:
         OnLButtonUp(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_RBUTTONDOWN:
         OnRButtonDown(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_RBUTTONUP:
         OnRButtonUp(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_LBUTTONDBLCLK:
         OnLButtonDoubleClick(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_MOUSEMOVE:
         OnMouseMove(wParam, LOWORD(lParam), HIWORD(lParam));
         break;
      case WM_NOTIFY:
         return OnNotify(wParam, (NMHDR *)lParam);
      }
      return DefWindowProc(WindowHandle, msg, wParam, lParam);
   }
   RECT GetClientRect()
   {
      RECT R;
      ::GetClientRect(WindowHandle, &R);
      return R;
   }
   HWND WindowHandle;

private:

   typedef std::map<HWND, WindowBaseClass *>WindowMapClass;
   static WindowMapClass WindowMap;
};

#endif
