#include "drawstd.h"
#include <utility>
#include <algorithm>
#include "image.h"
#include "math.h"

class ReaderBaseClass
{
public:
   virtual ~ReaderBaseClass() {}
   virtual bool Read(void *aDest, uint32_t aSize) = 0;
};

class FileReaderClass : public ReaderBaseClass
{
public:
   FileReaderClass(const char *aName)
   {
      f = fopen(aName, "rb"); // Will detect errors in read.
   }
   virtual ~FileReaderClass()
   {
      if(f)
         fclose(f);
   }
   bool Read(void *aDest, uint32_t aSize)
   {
      if(!f)
         return false;
      if(fread(aDest, 1, aSize, f) != aSize)
         return false;
      return true;
   }
private:
   FILE *f;
};

class MemReaderClass : public ReaderBaseClass
{
public:
   MemReaderClass(void *aPtr)
   {
      Ptr = (uint8_t *)aPtr;
   }
   bool Read(void *aDest, uint32_t aSize)
   {
      memcpy(aDest, Ptr, aSize);
      Ptr += aSize;
      return true;
   }
private:
   uint8_t *Ptr;
};

template <typename T> class BmpMaskImageClass
{
   struct BitMaskStruct
   {
      T Mask;
      T Shift;
      T Num;
   };
   static const BitMaskStruct DefaultBitMask[3];
public:
   static uint8_t *Read(ReaderBaseClass &aReader, int32_t &aWidth, int32_t &aHeight, uint32_t aCompression, uint32_t aBytesPixel, std::string &aErrorStr);
};

void ImageClass::Resize(int32_t aNewWidth, int32_t aNewHeight)
{
   int32_t y, x;
   uint8_t *n = new uint8_t [DrawMax(aNewHeight, Height) * (DrawMax(aNewWidth, Width) + 3) * 3];
   uint8_t *p = n;

   // We will do it in two steaps, first horz
   double Inc = 1.0*Width/aNewWidth;
   double d;

   for(y = 0; y < Height; y++)
   {
      for(x = 0, d = 0; x < aNewWidth; x++, d += Inc)
      {
         uint32_t l = DrawMin(int32_t(floor(d)), Width - 1);
         uint32_t m = DrawMin(int32_t(ceil(d)), Width - 1);
         double lw = 1.0 - (d - l);
         double mw = 1.0 - lw;

         for(int i = 0; i < 3; i++, p++)
         {
            uint16_t c = uint16_t(GetAt(y, l, i)*lw + GetAt(y, m, i)*mw);
            if(c > 255)
              c = 255;
            *p = c;
         }
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }

   // Then vert
   BytesPerLine = (3*aNewWidth + 3) & ~0x03;
   delete [] Data;
   Data = n;
   n = new uint8_t [aNewHeight * (aNewWidth + 3) * 3];
   p = n;

   Inc = 1.0*Height/aNewHeight;

   for(y = 0, d = 0; y < aNewHeight; y++, d += Inc)
   {
      uint32_t l = DrawMin(int32_t(floor(d)), Height - 1);
      uint32_t m = DrawMin(int32_t(ceil(d)), Height - 1);

      double lw = 1.0 - (d - l);
      double mw = 1.0 - lw;

      for(x = 0; x < aNewWidth; x++)
      {
         for(int i = 0; i < 3; i++, p++)
         {
            uint16_t c = uint16_t(GetAt(l, x, i)*lw + GetAt(m, x, i)*mw);
            if(c > 255)
              c = 255;
            *p = c;
         }
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }

   delete [] Data;
   Data = n;
   Width = aNewWidth;
   BytesPerLine = (3*Width + 3) & ~0x03;
   Height = aNewHeight;
}

void ImageClass::Mirror()
{
   int32_t y, x;
   uint8_t *n = new uint8_t [Height  * (Width + 3) * 3];
   uint8_t *p = n;

   for(y = 0; y < Height; y++)
   {
      for(x = 0; x < Width; x++, p += 3)
      {
         p[0] = GetAt(y, Width - x - 1, 0);
         p[1] = GetAt(y, Width - x - 1, 1);
         p[2] = GetAt(y, Width - x - 1, 2);
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   delete [] Data;
   Data = n;
}

void ImageClass::Crop(int32_t aLeft, int32_t aTop, int32_t aRight, int32_t aBottom)
{
   uint32_t NewWidth = aRight - aLeft;
   uint32_t NewHeight = aBottom - aTop;

   int32_t y, x;
   uint8_t *n = new uint8_t [NewHeight  * (NewWidth + 3) * 3];
   uint8_t *p = n;

   for(y = Height - aBottom; y < Height - aTop; y++)
   {
      for(x = aLeft; x < aRight; x++, p += 3)
      {
         p[0] = GetAt(y, x, 0);
         p[1] = GetAt(y, x, 1);
         p[2] = GetAt(y, x, 2);
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   delete [] Data;
   Data = n;
   Width = NewWidth;
   BytesPerLine = (3*Width + 3) & ~0x03;
   Height = NewHeight;
}

void ImageClass::Rotate()
{
   int32_t y, x;
   uint8_t *n = new uint8_t [Height  * (Width + 3) * 3];
   uint8_t *p = n;

   for(y = 0; y < Width; y++)
   {
      for(x = 0; x < Height; x++, p += 3)
      {
         p[0] = GetAt(x, Width - y - 1, 0);
         p[1] = GetAt(x, Width - y - 1, 1);
         p[2] = GetAt(x, Width - y - 1, 2);
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   delete [] Data;
   Data = n;
   std::swap(Height, Width);
   BytesPerLine = (3*Width + 3) & ~0x03;
}

void ImageClass::Move(int32_t aXOffset, int32_t aYOffset, COLORREF aColor)
{
   int32_t NewWidth = Width + aXOffset;
   int32_t NewHeight = Height + aYOffset;
   uint8_t *n = new uint8_t [NewHeight * (NewWidth + 3)*3];
   int32_t x, y;
   uint8_t *p = n;

   for(y = DrawMin(NewHeight, Height) - 1; y >= 0; y--)
   {
      for(x = 0; x < aXOffset; x++, p += 3)
      {
         p[0] = uint8_t(aColor >> 16);
         p[1] = uint8_t(aColor >>  8);
         p[2] = uint8_t(aColor >>  0);
      }
      for(x = 0; x < DrawMin(Width, NewWidth); x++, p += 3)
      {
         p[0] = GetAt(aYOffset >= 0 ? Height - y - 1: Height - y - 1 + aYOffset, aXOffset >= 0 ? x : x - aXOffset, 0);
         p[1] = GetAt(aYOffset >= 0 ? Height - y - 1: Height - y - 1 + aYOffset, aXOffset >= 0 ? x : x - aXOffset, 1);
         p[2] = GetAt(aYOffset >= 0 ? Height - y - 1: Height - y - 1 + aYOffset, aXOffset >= 0 ? x : x - aXOffset, 2);
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   for(y = 0; y < aYOffset; y++)
   {
      for(x = 0; x < NewWidth; x++, p += 3)
      {
         p[0] = uint8_t(aColor >> 16);
         p[1] = uint8_t(aColor >>  8);
         p[2] = uint8_t(aColor >>  0);
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   delete [] Data;
   Data = n;
   Width = NewWidth;
   BytesPerLine = (3*Width + 3) & ~0x03;
   Height = NewHeight;
}

bool ImageClass::Read(const char *aFileName)
{
   char Ext[256];
   _splitpath(aFileName, 0, 0, 0, Ext);
   if(!stricmp(Ext, ".bmp"))
   {
      FileReaderClass Reader(aFileName);
      BITMAPFILEHEADER BitmapFileHeader;
      ReadBitmapFileHeader(Reader, BitmapFileHeader);
      if(BitmapFileHeader.bfType != (('M' << 8) | 'B'))
      {
         ErrorStr = "Not a bitmap file";
         return false;
      }

      return ReadBmp(Reader);
   }
   else // It's JPG
   {
      int32_t x, y;
      uint8_t *T;
      uint32_t **JData;
      bool IsMono;
      if(!ReadJpegFile(aFileName, JData, Width, Height, IsMono, ErrorStr))
      {
         return false;
      }
      delete [] Data;
      T = Data = new uint8_t [3*(Width + 3)*(Height)];

      for(y = 0; y < Height; y++)
      {
         unsigned char *p;

         p = (unsigned char *)JData[Height - y - 1];

         if(IsMono)
         {
            for(x = 0; x < Width; x++, p ++, T += 3)
            {
               T[0] = p[0];
               T[1] = p[0];
               T[2] = p[0];
            }
         }
         else
         {
            for(x = 0; x < Width; x++, p += 3, T += 3)
            {
               T[0] = p[2];
               T[1] = p[1];
               T[2] = p[0];
            }
         }
         x = (unsigned int )T;
         x &= 0x03;
         T += (4 - x) & 0x03;
      }
      for(y = 0; y < Height; y++)
         delete [] JData[y];
      delete [] JData;
      BytesPerLine = (3*Width + 3) & ~0x03;
   }
   return true;
}

bool ImageClass::ReadBmp(ReaderBaseClass &aReader)
{
   BITMAPINFOHEADER BitmapInfoHeader;
   aReader.Read(&BitmapInfoHeader, sizeof(BitmapInfoHeader));

   Width = BitmapInfoHeader.biWidth;
   Height = BitmapInfoHeader.biHeight;
   BytesPerLine = (3*Width + 3) & ~0x03;
   delete [] Data;
   Data = 0;
   uint32_t NumColor = BitmapInfoHeader.biBitCount;

   if(NumColor == 24 || NumColor == 32)
   {
      Data = BmpMaskImageClass<uint32_t>::Read(aReader, Width, Height, BitmapInfoHeader.biCompression, NumColor/8, ErrorStr);
   }
   else if(NumColor == 16)
   {
      Data = BmpMaskImageClass<uint16_t>::Read(aReader, Width, Height, BitmapInfoHeader.biCompression, NumColor/8, ErrorStr);
   }
   else if(NumColor == 8 || NumColor == 4 || NumColor == 1)
   {
      Data = ReadBmpPalette(aReader, Width, Height, BitmapInfoHeader.biClrUsed, NumColor, ErrorStr);
   }
   else
   {
      ErrorStr = "Unknown bmp file format";
      return false;
   }
   return Data != 0;
}

uint8_t *ImageClass::ReadBmpPalette(ReaderBaseClass &aReader, int32_t &aWidth, int32_t aHeight, uint32_t aColors, uint32_t aBitPixel, std::string &ErrStr)
{
   uint32_t MaxColors = 1 << aBitPixel;
   uint32_t NumColors = aColors == 0 || aColors >= MaxColors ? MaxColors : aColors;

   uint32_t Color[256];
   if(!aReader.Read(Color, NumColors*sizeof(uint32_t)))
   {
      ErrStr = "Read error";
      return 0;
   }
   uint8_t *Data = new uint8_t [(Width + 3)*Height*3];
   uint8_t *Temp = new uint8_t [Height*BytesPerLine];
   if(!aReader.Read(Temp, Height*BytesPerLine))
   {
      ErrStr = "Read error";
      delete [] Data;
      Data = 0;
      delete [] Temp;
      return 0;
   }
   int32_t i, j;
   uint8_t *d, *p;

   for(i = 0, d = Data, p = Temp; i < Height; i++)
   {
      uint32_t BitOffset = 0;
      for(j = 0; j < Width; j++, d += 3)
      {
         uint32_t idx = *p;
         uint32_t mask = ((1 << aBitPixel) - 1) << BitOffset;
         idx &= mask;
         idx >>= BitOffset;
         d[0] = uint8_t((Color[idx] >>  0) & 0xFF);
         d[1] = uint8_t((Color[idx] >>  8) & 0xFF);
         d[2] = uint8_t((Color[idx] >> 16) & 0xFF);
         BitOffset += aBitPixel;
         if(BitOffset >= 8)
         {
            BitOffset = 0;
            p++;
         }
      }
      j = ((unsigned int )d) & 0x03;
      d += (4 - j) & 0x03;
      j = ((unsigned int )p) & 0x03;
      p += (4 - j) & 0x03;
   }
   delete [] Temp;
   return Data;
}

bool ImageClass::WriteJpg(const char *aFileName, int aQuality)
{
   uint32_t **JData;
   JData = new uint32_t * [Height];
   int32_t y, x;

   uint8_t *t = Data;
   for(y = 0; y < Height; y++)
   {
      JData[Height - y - 1] = new uint32_t [Width];
      uint8_t *p = (uint8_t *)JData[Height - y - 1];
      for(x = 0; x < Width; x++, p += 3, t += 3)
      {
         p[0] = t[2];
         p[1] = t[1];
         p[2] = t[0];
      }
      t = (uint8_t *)((((uint32_t )t) + 3) & ~0x03);
   }
   WriteJpegFile(aFileName, JData, Width, Height, aQuality);
   for(y = 0; y < Height; y++)
      delete [] JData[y];
   delete [] JData;

   return TRUE;
}

bool ImageClass::WriteBmp(const char *aFileName)
{
   BITMAPFILEHEADER BitmapFileHeader;
   BITMAPINFOHEADER BitmapInfoHeader;
   uint32_t BytesPerLine = (3*Width + 3) & ~0x03;

   BitmapFileHeader.bfType = (('M' << 8) | 'B');
   BitmapFileHeader.bfSize = sizeof(BitmapFileHeader) + sizeof(BitmapInfoHeader) + BytesPerLine*Height;
   BitmapFileHeader.bfReserved1 = 0;
   BitmapFileHeader.bfReserved2 = 0;
   BitmapFileHeader.bfOffBits = 0x36;

   InitializBitmapInfoHeader(BitmapInfoHeader);

   FILE *f = fopen(aFileName, "wb");
   if(!f)
      return FALSE;

   WriteBitmapFileHeader(f, BitmapFileHeader);
   fwrite(&BitmapInfoHeader, sizeof(BitmapInfoHeader), 1, f);
   fwrite(Data, BytesPerLine*Height, 1, f);
   fclose(f);

   return true;
}

void ImageClass::InitializBitmapInfoHeader(BITMAPINFOHEADER &aBitmapInfoHeader)
{
   aBitmapInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
   aBitmapInfoHeader.biWidth = Width;
   aBitmapInfoHeader.biHeight = Height;
   aBitmapInfoHeader.biPlanes = 1;
   aBitmapInfoHeader.biBitCount = 24;
   aBitmapInfoHeader.biCompression = BI_RGB;
   aBitmapInfoHeader.biSizeImage = 0;
   aBitmapInfoHeader.biXPelsPerMeter = 0;
   aBitmapInfoHeader.biYPelsPerMeter = 0;
   aBitmapInfoHeader.biClrUsed = 0;
   aBitmapInfoHeader.biClrImportant = 0;
}

HANDLE ImageClass::CreateClipboardData()
{
   BITMAPINFOHEADER BitmapInfoHeader;
   InitializBitmapInfoHeader(BitmapInfoHeader);

   uint32_t BytesPerLine = (3*Width + 3) & ~0x03;

   HGLOBAL HData = ::GlobalAlloc(GMEM_DDESHARE, sizeof(BITMAPINFOHEADER) + BytesPerLine*Height);
   uint8_t *p = (uint8_t *)::GlobalLock(HData);
   memcpy(p, &BitmapInfoHeader, sizeof(BitmapInfoHeader));
   memcpy(p + sizeof(BitmapInfoHeader), Data, BytesPerLine*Height);

   return HData;
}

template <typename T>
uint8_t *BmpMaskImageClass<T>::Read(ReaderBaseClass &aReader, int32_t &aWidth, int32_t &aHeight, uint32_t aCompression, uint32_t aBytesPixel, std::string &aErrStr)
{
   BitMaskStruct BitMask[3];

   if(aCompression == BI_RGB)
   {
      memcpy(&BitMask, DefaultBitMask, sizeof(BitMask));
   }
   else if(aCompression == BI_BITFIELDS)
   {
      size_t i, j;
      for(i = 0; i < 3; i++)
      {
         uint32_t Temp;
         if(!aReader.Read(&Temp, sizeof(Temp)))
         {
            aErrStr = "Read Error";
            return 0;
         }
         BitMask[i].Mask = T(Temp);
         if(BitMask[i].Mask == 0)
         {
            BitMask[i].Shift = 0;
            BitMask[i].Num = 0;
         }
         else
         {
            for(j = 0; (BitMask[i].Mask & (0x01 << j)) == 0; j++)
            { /* Just counting */ }
            BitMask[i].Shift = T(j);
            for(; (BitMask[i].Mask & (0x01 << j)) != 0 && j < 8*sizeof(T); j++)
            { /* Just counting */ }
            BitMask[i].Num = T(j - BitMask[i].Shift);
         }
      }
   }
   else
   {
      return 0;
   }
   uint8_t *Data = new uint8_t [(aWidth + 3)*aHeight*3];
   uint32_t BytesPerLine = (aBytesPixel*aWidth + 3) & ~0x03;
   uint8_t *Temp = new uint8_t [BytesPerLine*aHeight + 3];

   if(!aReader.Read(Temp, BytesPerLine*aHeight))
   {
      aErrStr = "Read Error";
      delete [] Data;
      Data = 0;
      delete [] Temp;
      return 0;
   }
   int32_t i, j;
   uint8_t *p, *d;
   for(i = 0, p = Temp, d = Data; i < aHeight; i++)
   {
      for(j = 0; j < aWidth; j++, p += aBytesPixel, d += 3)
      {
         T N = *((T *)p);
         d[2] = (uint8_t )(((N & BitMask[0].Mask) >> BitMask[0].Shift) << (8 - BitMask[0].Num));
         d[1] = (uint8_t )(((N & BitMask[1].Mask) >> BitMask[1].Shift) << (8 - BitMask[1].Num));
         d[0] = (uint8_t )(((N & BitMask[2].Mask) >> BitMask[2].Shift) << (8 - BitMask[2].Num));
      }
      j = ((uint32_t )d) & 0x03;
      d += (4 - j) & 0x03;
      j = ((uint32_t )p) & 0x03;
      p += (4 - j) & 0x03;
   }
   delete [] Temp;
   return Data;
}

bool ImageClass::CreateFromClipboard(void *aPtr)
{
   MemReaderClass Reader(aPtr);

   return ReadBmp(Reader);
}

bool ImageClass::CreateFromScreenDump()
{
   delete [] Data;
   Data = 0;
   HDC dc = GetDC(0);
   if(!dc)
   {
      ErrorStr = std::string("GetDC Failed: ") + ToString(GetLastError());
      return false;
   }
   HDC MemDc = CreateCompatibleDC(dc);
   if(!MemDc)
   {
      ErrorStr = std::string("CreateCompatibleDC Failed: ") + ToString(GetLastError());
      return false;
   }

   Width = GetDeviceCaps(dc, HORZRES);
   Height = GetDeviceCaps(dc, VERTRES);
   BytesPerLine = (3*Width + 3) & ~0x03;

   HBITMAP BitMap = CreateCompatibleBitmap(dc, Width, Height);
   if(!BitMap)
   {
      ErrorStr = "CreateCompatibleBitmap error: ";
      ErrorStr += ToString(GetLastError());
      return false;
   }
   HGDIOBJ OldBitmap = SelectObject(MemDc, BitMap);
   if(!BitBlt(MemDc, 0, 0, Width, Height, dc, 0, 0, SRCCOPY))
   {
      ErrorStr = std::string("BitBlt Failed: ") + ToString(GetLastError());
      return false;
   }
   BITMAP Bm;
   GetObject(BitMap, sizeof(Bm), &Bm);
   char *Buffer = new char [Height*Width*4];

   BITMAPINFOHEADER BitmapInfoHeader;
   BitmapInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
   BitmapInfoHeader.biWidth = Width;
   BitmapInfoHeader.biHeight = Height;
   BitmapInfoHeader.biPlanes = 1;
   BitmapInfoHeader.biBitCount = Bm.bmBitsPixel;
   BitmapInfoHeader.biCompression = BI_RGB;
   BitmapInfoHeader.biSizeImage = 0;
   BitmapInfoHeader.biXPelsPerMeter = 0;
   BitmapInfoHeader.biYPelsPerMeter = 0;
   BitmapInfoHeader.biClrUsed = 0;
   BitmapInfoHeader.biClrImportant = 0;

   SelectObject(MemDc, OldBitmap);

   if(GetDIBits(MemDc,
                BitMap,
                0,
                Height,
                Buffer,
                (BITMAPINFO *)&BitmapInfoHeader,
                DIB_RGB_COLORS) != Height)
   {
      ErrorStr = std::string("GetDIBits Failed: ") + ToString(GetLastError());
      return false;
   }

   MemReaderClass MemReader(Buffer);

   if(Bm.bmBitsPixel == 24 || Bm.bmBitsPixel == 32)
   {
      Data = BmpMaskImageClass<uint32_t>::Read(MemReader, Width, Height, BitmapInfoHeader.biCompression, Bm.bmBitsPixel/8, ErrorStr);
   }
   else if(Bm.bmBitsPixel == 16)
   {
      Data = BmpMaskImageClass<uint16_t>::Read(MemReader, Width, Height, BitmapInfoHeader.biCompression, Bm.bmBitsPixel/8, ErrorStr);
   }
   else if(Bm.bmBitsPixel == 8 || Bm.bmBitsPixel == 4 || Bm.bmBitsPixel == 1)
   {
      Data = ReadBmpPalette(MemReader, Width, Height, BitmapInfoHeader.biClrUsed, Bm.bmBitsPixel, ErrorStr);
   }
   else
   {
      ErrorStr = "Unknown screen format";
   }
   delete [] Buffer;
   DeleteObject(BitMap);
   DeleteDC(MemDc);
   ReleaseDC(0, dc);
   return Data != 0;
}

bool ImageClass::operator == (const ImageClass &rhs)
{
   if(Width != rhs.Width || Height != rhs.Height)
      return false;
   int32_t x, y;
   uint8_t *p = Data;
   for(y = 0; y < Height; y++)
   {
      for(x = 0; x < Width; x++, p+= 3)
      {
         if(GetAt(y, x, 0) != rhs.GetAt(y, x, 0) ||
            GetAt(y, x, 1) != rhs.GetAt(y, x, 1) ||
            GetAt(y, x, 2) != rhs.GetAt(y, x, 2))
         {
            return false;
         }
      }
      p = (uint8_t *)((((uint32_t )p) + 3) & ~0x03);
   }
   return true;
}

bool ImageClass::operator != (const ImageClass &rhs)
{
   return !(*this == rhs);
}

void ImageClass::WriteBitmapFileHeader(FILE *aFile, BITMAPFILEHEADER &aBitmapFileHeader)
{
   fwrite((void *)&aBitmapFileHeader.bfType, sizeof(WORD), 1, aFile);
   fwrite((void *)&aBitmapFileHeader.bfSize, sizeof(DWORD), 1, aFile);
   fwrite((void *)&aBitmapFileHeader.bfReserved1, sizeof(WORD), 1, aFile);
   fwrite((void *)&aBitmapFileHeader.bfReserved2, sizeof(WORD), 1, aFile);
   fwrite((void *)&aBitmapFileHeader.bfOffBits, sizeof(DWORD), 1, aFile);
}

void ImageClass::ReadBitmapFileHeader(ReaderBaseClass &aReader, BITMAPFILEHEADER &aBitmapFileHeader)
{
   aReader.Read((void *)&aBitmapFileHeader.bfType, sizeof(WORD));
   aReader.Read((void *)&aBitmapFileHeader.bfSize, sizeof(DWORD));
   aReader.Read((void *)&aBitmapFileHeader.bfReserved1, sizeof(WORD));
   aReader.Read((void *)&aBitmapFileHeader.bfReserved2, sizeof(WORD));
   aReader.Read((void *)&aBitmapFileHeader.bfOffBits, sizeof(DWORD));
}

template<> const BmpMaskImageClass<uint16_t>::BitMaskStruct BmpMaskImageClass<uint16_t>::DefaultBitMask[3] =
{
   {0x7C00, 10, 5},
   {0x03E0,  5, 5},
   {0x001F,  0, 5}
};

template<> const BmpMaskImageClass<uint32_t>::BitMaskStruct BmpMaskImageClass<uint32_t>::DefaultBitMask[3] =
{
   {0xFF0000, 16, 8},
   {0x00FF00,  8, 8},
   {0x0000FF,  0, 8}
};

//--- EOF
