#include "imagewin.h"
#include "main.h"
#include "drawres.h"
#include "sizedialog.h"
#include <math.h>

BOOL ImageWindowClass::Create(DrawAppClass *aParent, uint32_t aWidth, uint32_t aHeight)
{
   Parent = aParent;
   WNDCLASS wc;
   memset(&wc, 0, sizeof(WNDCLASS));
   wc.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
   wc.lpfnWndProc = (WNDPROC )&ImageWindowClass::WndProcEntry;
   wc.hInstance = AppClass::GetApp()->GetInstanceHandle();
   wc.hbrBackground = (HBRUSH )(COLOR_WINDOW + 1);
   wc.lpszClassName = "ImageWndClass";
   wc.lpszMenuName = MAKEINTRESOURCE(IDMAINMENU);
   wc.hCursor = LoadCursor(NULL,IDC_ARROW);
   wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);
   if (!RegisterClass(&wc))
     return FALSE;
   WindowHandle = CreateWindow("ImageWndClass",
                               "Draw",
                               WS_VISIBLE | WS_CLIPSIBLINGS | WS_CHILD,
                               0, 0, aWidth, aHeight,
                               Parent->GetWindowHandle(),
                               0,
                               AppClass::GetApp()->GetInstanceHandle(),
                               (LPVOID )this);

   SetScroll(aWidth, aHeight, GetDispWidth(), GetDispHeight());

   if(Parent)
      Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
   return TRUE;
}

void ImageWindowClass::MoveWindow(RECT &aR, bool aReDraw)
{
   Zoom = NormalZoom;
   SetScroll(aR.right - aR.left, aR.bottom - aR.top, GetDispWidth(), GetDispHeight());
   ::MoveWindow(WindowHandle, aR.left, aR.top, aR.right, aR.bottom, aReDraw);
   if(Parent)
      Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
}

void ImageWindowClass::SetImage(ImageClass *aImage)
{
   delete CurrentImage;
   CurrentImage = aImage;
   Zoom = NormalZoom;

   RECT R = GetClientRect();
   SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
   InvalidateRect(WindowHandle, 0, TRUE);
   if(Parent)
      Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
}

void ImageWindowClass::UpdateImage()
{
   Zoom = NormalZoom;
   RECT R = GetClientRect();
   SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
   InvalidateRect(WindowHandle, 0, TRUE);
   if(Parent)
      Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
}

void ImageWindowClass::SetZoom()
{
   SizeDialogClass SizeDialog(WindowHandle, false);
   if(SizeDialog.Run(GetImage()->GetWidth(), GetImage()->GetHeight()))
   {
      Zoom = SizeDialog.GetX()*NormalZoom/CurrentImage->GetWidth();
      RECT R = GetClientRect();
      SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
      InvalidateRect(WindowHandle, 0, TRUE);
      if(Parent)
         Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
   }
}

void ImageWindowClass::ZoomToFit()
{
   RECT R = GetClientRect();
   double ZoomX = R.right*NormalZoom/CurrentImage->GetWidth();
   double ZoomY = R.bottom*NormalZoom/CurrentImage->GetHeight();
   Zoom = DrawMin(ZoomX, ZoomY);
   SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
   InvalidateRect(WindowHandle, 0, TRUE);
   if(Parent)
      Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
}

void ImageWindowClass::ShowImage(HDC aDC)
{
   if(!CurrentImage)
      return;
   BITMAPINFOHEADER bmiHeader;
   CurrentImage->InitializBitmapInfoHeader(bmiHeader);
   SetStretchBltMode(aDC, HALFTONE);
   StretchDIBits(aDC,
                 -XOffset + DispOffsetX, -YOffset + DispOffsetY, GetDispWidth(), GetDispHeight(),
                 0, 0, bmiHeader.biWidth, bmiHeader.biHeight,
                 CurrentImage->GetData(),
                 (LPBITMAPINFO)&bmiHeader, DIB_RGB_COLORS, SRCCOPY);
}

bool ImageWindowClass::OnCommand(int aId, HWND, UINT)
{
   switch(aId)
   {
   case IDM_ZOOM_IN:
      if(Zoom < 20 && CurrentImage)
      {
         Zoom *= 1.25;
         if(Zoom >= 0.95 && Zoom <= 1.05)
            Zoom = 1.0;
         RECT R = GetClientRect();
         SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
         InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
         if(Parent)
            Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
      }
      break;
   case IDM_ZOOM_OUT:
      if(Zoom > 0.02 && CurrentImage)
      {
         Zoom *= 1/1.25;
         if(Zoom >= 0.95 && Zoom <= 1.05)
            Zoom = 1.0;
         RECT R = GetClientRect();
         SetScroll(R.right, R.bottom, GetDispWidth(), GetDispHeight());
         InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
         if(Parent)
            Parent->SetZoomSize(GetDispWidth(), GetDispHeight());
      }
      break;
   case IDM_SCROLL_LINE_DOWN:
      OnVScroll(SB_LINEDOWN, 0);
      InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
      break;
   case IDM_SCROLL_LINE_UP:
      OnVScroll(SB_LINEUP, 0);
      InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
      break;
   case IDM_SCROLL_LINE_LEFT:
      OnHScroll(SB_LINELEFT, 0);
      InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
      break;
   case IDM_SCROLL_LINE_RIGHT:
      OnHScroll(SB_LINERIGHT, 0);
      InvalidateRect(WindowHandle, 0, TRUE); // Will erase background, and paint image
      break;
   case IDM_RESIZE:
      Resize();
      break;
   case IDM_MIRROR:
      Mirror();
      break;
   case IDM_ROTATE:
      Rotate();
      break;
   case IDM_SET_ZOOM:
      SetZoom();
      break;
   case IDM_ZOOM_TO_FIT:
      ZoomToFit();
      break;
   case IDM_CROP:
      Crop();
      break;
   default:
      return FALSE;
   }
   return TRUE;
}

int32_t ImageWindowClass::GetDispWidth(void)
{
   if(CurrentImage)
      return int32_t(CurrentImage->GetWidth()*Zoom/NormalZoom);

   return 0;
}

int32_t ImageWindowClass::GetDispHeight(void)
{
   if(CurrentImage)
      return int32_t(CurrentImage->GetHeight()*Zoom/NormalZoom);
   return 0;
}

LRESULT ImageWindowClass::OnPaint(void)
{
   PAINTSTRUCT PaintStruct;

   HBRUSH BackBrush;
   HDC DCHandle = BeginPaint(WindowHandle, &PaintStruct);
   if(CurrentImage)
      ShowImage(DCHandle);
   /* Erase space to the right and at the buttom */
   RECT Rect, ClientRect = GetClientRect();
   BackBrush = CreateSolidBrush(BackColor);

   if(GetDispWidth() + DispOffsetX < ClientRect.right)
   {
      Rect.left = GetDispWidth() + DispOffsetX;
      Rect.right = ClientRect.right;
      Rect.top = 0;
      Rect.bottom = ClientRect.bottom;
      FillRect(DCHandle, &Rect, BackBrush);
   }

   if(GetDispHeight() + DispOffsetY < ClientRect.bottom)
   {
      Rect.left = 0;
      Rect.right = ClientRect.right;
      Rect.top = GetDispHeight() + DispOffsetY;
      Rect.bottom = ClientRect.bottom;
      FillRect(DCHandle, &Rect, BackBrush);
   }
   if(DispOffsetX)
   {
      Rect.left = 0;
      Rect.right = DispOffsetX;
      Rect.top = 0;
      Rect.bottom = ClientRect.bottom;
      FillRect(DCHandle, &Rect, BackBrush);
   }
   if(DispOffsetY)
   {
      Rect.left = 0;
      Rect.right = ClientRect.right;
      Rect.top = 0;
      Rect.bottom = DispOffsetY;
      FillRect(DCHandle, &Rect, BackBrush);
   }

   DeleteObject(BackBrush);
   if(IsRect)
   {
      FrameRect = FrameRectFromImageRect(ImageRect);
      DrawFocusRect(DCHandle, &FrameRect);
   }
   EndPaint(WindowHandle, &PaintStruct);

   return FALSE;
}

void ImageWindowClass::Resize()
{
   SizeDialogClass ResizeDialog(WindowHandle);
   if(ResizeDialog.Run(GetImage()->GetWidth(), GetImage()->GetHeight()))
   {
      WaitCursorClass WaitCursor;
      GetImage()->Resize(ResizeDialog.GetX(), ResizeDialog.GetY());
      UpdateImage();
      Parent->SetImageSize(GetImage()->GetWidth(), GetImage()->GetHeight());
   }
}

void ImageWindowClass::Mirror()
{
   WaitCursorClass WaitCursor;
   GetImage()->Mirror();
   UpdateImage();
}

void ImageWindowClass::Rotate()
{
   WaitCursorClass WaitCursor;
   GetImage()->Rotate();
   UpdateImage();
}

void ImageWindowClass::Crop()
{
   if(IsRect)
   {
      WaitCursorClass WaitCursor;
      GetImage()->Crop(ImageRect.left, ImageRect.top, ImageRect.right, ImageRect.bottom);
      RemoveRect();
      UpdateImage();
      Parent->SetImageSize(GetImage()->GetWidth(), GetImage()->GetHeight());
   }
}

void ImageWindowClass::OnLButtonDown(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
{
   if(IsRect && (aFlags & MK_CONTROL))
   {
      int32_t mx = (FrameRect.right - FrameRect.left)/2 + FrameRect.left;
      int32_t my = (FrameRect.bottom - FrameRect.top)/2 + FrameRect.top;
      if(aXPos > mx)
      {
         if(aYPos > my)
         {
            FrameStart.x = FrameRect.left;
            FrameStart.y = FrameRect.top;
         }
         else
         {
            FrameStart.x = FrameRect.left;
            FrameStart.y = FrameRect.bottom;
         }
      }
      else
      {
         if(aYPos > my)
         {
            FrameStart.x = FrameRect.right;
            FrameStart.y = FrameRect.top;
         }
         else
         {
            FrameStart.x = FrameRect.right;
            FrameStart.y = FrameRect.bottom;
         }
      }
      RemoveRect();
      CalcFrameRect(aXPos, aYPos);
      HDC dc = GetDC(WindowHandle);
      DrawFocusRect(dc, &FrameRect);
      IsRect = true;
      ReleaseDC(WindowHandle, dc);
      Parent->SetFrameSize(ImageRect.right - ImageRect.left, ImageRect.bottom - ImageRect.top);
      IsFraming = true;
   }
   else if(aFlags & MK_SHIFT)
   {
      RemoveRect();
      UpdateImage();
      IsMoving = true;
      DispOffsetX = 0;
      DispOffsetY = 0;
      MovingX = aXPos;
      MovingY = aYPos;
   }
   else
   {
      RemoveRect();
      FrameStart.x = aXPos;
      FrameStart.y = aYPos;
      IsFraming = true;
   }
}

void ImageWindowClass::OnLButtonUp(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
{
   if(IsFraming)
   {
      RemoveRect();
      CalcFrameRect(aXPos, aYPos);

      ImageRect = ImageRectFromFrameRect(FrameRect);

      if(ImageRect.bottom > GetImage()->GetHeight())
         ImageRect.bottom = GetImage()->GetHeight();
      if(ImageRect.right > GetImage()->GetWidth())
         ImageRect.right = GetImage()->GetWidth();

      HDC dc = GetDC(WindowHandle);
      DrawFocusRect(dc, &FrameRect);
      IsRect = true;
      ReleaseDC(WindowHandle, dc);
      Parent->SetFrameSize(ImageRect.right - ImageRect.left, ImageRect.bottom - ImageRect.top);
      IsFraming = false;
   }
   else if(IsMoving)
   {
      DispOffsetX = aXPos - MovingX;
      DispOffsetY = aYPos - MovingY;
      GetImage()->Move(DispOffsetX, DispOffsetY, BackColor);
      DispOffsetX = 0;
      DispOffsetY = 0;
      InvalidateRect(WindowHandle, 0, TRUE);
      IsMoving = false;
      UpdateImage();
      Parent->SetImageSize(GetImage()->GetWidth(), GetImage()->GetHeight());
   }
}

void ImageWindowClass::OnMouseMove(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
{
   if(IsFraming)
   {
      RemoveRect();
      CalcFrameRect(aXPos, aYPos);
      ImageRect = ImageRectFromFrameRect(FrameRect);

      if(ImageRect.bottom > GetImage()->GetHeight())
         ImageRect.bottom = GetImage()->GetHeight();
      if(ImageRect.right > GetImage()->GetWidth())
         ImageRect.right = GetImage()->GetWidth();

      HDC dc = GetDC(WindowHandle);
      DrawFocusRect(dc, &FrameRect);
      IsRect = true;
      ReleaseDC(WindowHandle, dc);
      Parent->SetFrameSize(ImageRect.right - ImageRect.left, ImageRect.bottom - ImageRect.top);
   }
   else if(IsMoving)
   {
      DispOffsetX = aXPos - MovingX;
      DispOffsetY = aYPos - MovingY;
      InvalidateRect(WindowHandle, 0, TRUE);
   }
}

void ImageWindowClass::CalcFrameRect(int32_t aX, int32_t aY)
{
   FrameRect.left = DrawMin<int32_t>(aX, FrameStart.x);
   FrameRect.right = DrawMax<int32_t>(aX, FrameStart.x);
   FrameRect.top = DrawMin<int32_t>(aY, FrameStart.y);
   FrameRect.bottom = DrawMax<int32_t>(aY, FrameStart.y);
}

void ImageWindowClass::OnRButtonDown(uint32_t aFlags, int16_t aXPos, int16_t aYPos)
{
   RemoveRect();
}

void ImageWindowClass::RemoveRect()
{
   if(IsRect)
   {
      HDC dc = GetDC(WindowHandle);
      DrawFocusRect(dc, &FrameRect);
      IsRect = false;
      ReleaseDC(WindowHandle, dc);
      Parent->SetFrameSize(0, 0);
   }
}

RECT ImageWindowClass::ImageRectFromFrameRect(RECT &aRect)
{
   RECT R = aRect;
   R.left += XOffset;
   R.right += XOffset;
   R.top += YOffset;
   R.bottom += YOffset;

   R.left /= Zoom;
   R.right /= Zoom;
   R.top /= Zoom;
   R.bottom /= Zoom;
   return R;
}

RECT ImageWindowClass::FrameRectFromImageRect(RECT &aRect)
{
   RECT R = aRect;
   R.left *= Zoom;
   R.right *= Zoom;
   R.top *= Zoom;
   R.bottom *= Zoom;

   R.left -= XOffset;
   R.right -= XOffset;
   R.top -= YOffset;
   R.bottom -= YOffset;

   return R;
}

